/*
Filename:	decodeevt.c

Purpose:	Decodes event keyword in the input science event file,generates
		TIME,PCUID,ANODE,PROPANE,and CHANNEL columns in output file.

Author:		Zhiyu Guo

Date:		May 1998

History:

*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include "fitsio.h"
#include "cfortran.h"
#include "ftools.h"
#include "ftoolstruct.h"
#include "xte.h"
#include "xpi.h"

#define STR_KEY_LEN 256	/*Define the size of string value in a keyword */
#define SPEC_CHAN 129	/*Define the number of spectral channel in SA files */
#define NUM_FIELDS 30	/*Define the number of anode wire layers in PCUs */

int evt_rd_param();
int getindex(char [],char []);
void getinfo(int, int *, char *); 

void decodeevt()
{
    	fitsfile *fptr;    /* pointer to the FITS file, defined in fitsio.h */
	fitsfile *outptr;
    	int status, hdutype, anynull, hdu_num,morekeys,tfields;
	int event[2], evt[3],pcuid,cbitnum,temp,temp1,pos,propane;
	char anode[16],tmpstr[2];
	char * tmpanode;
	long frow, felem,nelem,nrows,naxes2,longnull,i,k;
	double time1, dblnull;
	char nfound[STR_KEY_LEN],infile[STR_KEY_LEN],outfile[STR_KEY_LEN];
	char keyname[30];

/* Define keyword values for the output file */

	char extname[]="XTE_SE";
	char *ttype[]={"TIME", "CHANNEL", "PCUID", "ANODE"};
	char *tform[]={"D", "I", "I", "15A"};
	char *tunit[]={"s", "channel", "pcuid", "layer"};

        char *ttype1[]={"TIME", "CHANNEL", "PCUID", "ANODE", "PROPANE"};
        char *tform1[]={"D", "I", "I", "15A","I"};
        char *tunit1[]={"s", "channel", "pcuid", "layer","propane"};

	char comm[]=" ";
	char **tevtb;
	char tevtb_out[1440];
	char tddes[67];

	status = 0;
	
printf("\n\nRunning DECODEEVT version 1.0\n");
printf("================================================\n");

status=evt_rd_param(infile,outfile);
if(status !=0){
	printf("\n\nCould not complete evt_rd_param call\n");
	printf("Terminate run\n");
	exit(1);
} 
	
/* Open input file */

    fits_open_file(&fptr, infile, READONLY, &status);

/* Create an empty file for writing to */

	fits_create_file(&outptr,outfile,&status);
	if(status != 0){
	printf("\n Can't create the output file, it may already exist ! \n\n");
	exit(1);
	}

	morekeys=0;

/* Copy the primary header of the input file to the file just created */

	fits_copy_hdu(fptr, outptr, morekeys, &status);

fits_write_chksum(outptr,&status);

/* Now move to the first extension of the input file */

	hdu_num=2;
	fits_movabs_hdu(fptr,hdu_num,&hdutype,&status);

/* Read 'NAXIS2' value to determine the number of rows */

	fits_read_key(fptr,TSTRING,"TDDES2",tddes,nfound,&status);
	fits_read_key(fptr,TSTRING,"DATAMODE",keyname,nfound,&status);
printf("\nDATAMODE = %s\n",keyname);

tevtb=(char**) calloc(21,sizeof(char*));
for(i=0;i<21;i++){
        tevtb[i]=(char*) calloc(67,sizeof(char));
}

	fits_read_key_longstr(fptr,"TEVTB2",tevtb,comm,&status);
	strcpy(tevtb_out,tevtb[0]);

        fits_read_key(fptr, TLONG, "NAXIS2",  &naxes2,nfound,&status);
printf("\nNAXIS2 = %ld\n",naxes2);

frow=1;
felem=1;
nelem=1;
longnull=0;
dblnull=0;

if((strncmp(keyname,"E_1ms_128",9) == 0)){  /*The 1st IF statement*/

/* Read EVENT column from the binary extension of the input file, select rows
   that have M token value of 1, i.e., the first element in each row of EVENT
   colum should be greater than 127.  Record the total number of such rows,
   this number will be used as the row number in the output file.
*/

k=0;
for(i=0;i<naxes2;i++){
	fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);
	if(event[0] > 127){
		k++;
	}
}
nrows=k;
printf("\nThere are %ld events corresponding to M token value equal to 1.\n",nrows);

/* Create a binary extension in the output file created above */
 
tfields=4;
 
fits_create_tbl(outptr,BINARY_TBL,nrows,tfields,ttype,tform,tunit,
extname,&status);

/* Read column values from the first extension of the input file and write
   them to a column in the binary extension of the output file created above */

tmpanode = (char*) malloc(16*sizeof(char));

k=0;
for(i=0;i<naxes2;i++){
	fits_read_col(fptr, TDOUBLE, 1, i+1, felem, 1,
                &dblnull, &time1, &anynull, &status);
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);

	if(event[0] > 127){
		event[0]=event[0]-127; /*ignore 1st bit (M token) */
		event[1]=(event[1] >> 3); /*ignore last 3 bits (time bits) */
		getinfo(event[1],&pcuid,anode); 
		strcpy(tmpanode,anode);
		fits_write_col(outptr,TDOUBLE,1,k+1,felem,1,&time1,&status);
		fits_write_col(outptr,TINT,2,k+1,felem,1,&event[0],&status);
		fits_write_col(outptr,TINT,3,k+1,felem,1,&pcuid,&status);
		fits_write_col(outptr,TSTRING,4,k+1,felem,1,&tmpanode,&status);
		k++;
	}
}

fits_modify_comment(outptr,"TTYPE2",
	"label for field   2 (relative channels)",&status);

}       /*end of the 1st IF statement*/

if((strncmp(keyname,"E_500us_64",10) == 0)){  /*The 2nd IF statement*/

/* Read EVENT column from the binary extension of the input file, select rows
   that have M token value of 1, i.e., the first element in each row of EVENT
   colum should be greater than 127.  Record the total number of such rows,
   this number will be used as the row number in the output file.
*/
 
k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);
        if(event[0] > 127){
                k++;
        }
}
nrows=k;
printf("\nThere are %ld events corresponding to M token value equal to 1.\n",nrows);
 
/* Create a binary extension in the output file created above */
 
tfields=4;
 
fits_create_tbl(outptr,BINARY_TBL,nrows,tfields,ttype,tform,tunit,
extname,&status);
 
/* Read column values from the first extension of the input file and write
   them to a column in the binary extension of the output file created above */
 
tmpanode = (char*) malloc(16*sizeof(char));
 
k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TDOUBLE, 1, i+1, felem, 1,
                &dblnull, &time1, &anynull, &status);
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);

        if(event[0] > 127 ){
		temp = event[0] >> 1;  /*get rid of the last bit so that the
                                   remaining bits only have infomation
                                   about M and C tokens */
 
        	temp1=event[0]-(temp*2); /*get the last bit in event[0] */

		event[0]=(event[0] >> 1) - 64;  /*make sure it is a real 
						event (Mtoken value
                                   equals 1), and get rid of the 1st bit (M)
                                   so that the remaining bits only have
                                   information about channels (C token)  */

		event[1]=temp1*16+(event[1]>>4); /* get the 5 bits in
							   Z token */

                getinfo(event[1],&pcuid,anode); 
                strcpy(tmpanode,anode);
                fits_write_col(outptr,TDOUBLE,1,k+1,felem,1,&time1,&status);
                fits_write_col(outptr,TINT,2,k+1,felem,1,&event[0],&status);
                fits_write_col(outptr,TINT,3,k+1,felem,1,&pcuid,&status);
                fits_write_col(outptr,TSTRING,4,k+1,felem,1,&tmpanode,&status);
                k++;
        }
}
 
fits_modify_comment(outptr,"TTYPE2",
        "label for field   2 (relative channels)",&status);

}  /*end of the 2nd IF */

if((strncmp(keyname,"E_1us_1",7) == 0) || (strncmp(keyname,"E_2us_2",7) == 0) || (strncmp(keyname,"E_4us_4",7) == 0) || (strncmp(keyname,"E_8us_8",7) == 0) || (strncmp(keyname,"E_16us_16",9) == 0) || (strncmp(keyname,"E_31us_16",9) == 0) || (strncmp(keyname,"E_62us_32",9) == 0) || (strncmp(keyname,"E_62us_64",9) == 0) || (strncmp(keyname,"E_125us_64",10) == 0) || (strncmp(keyname,"E_250us_128",11) == 0)){  /*The 3rd IF statement*/
 

/* Read EVENT column from the binary extension of the input file, select rows
   that have M token value of 1, i.e., the first element in each row of EVENT
   colum should be greater than 127.  Record the total number of such rows,
   this number will be used as the row number in the output file.
*/
 
k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);
        if(event[0] > 127){
                k++;
        }
}
nrows=k;
printf("\nThere are %ld events corresponding to M token value equal to 1.\n",nrows);
 
/* Create a binary extension in the output file created above */
 
tfields=4;
fits_create_tbl(outptr,BINARY_TBL,nrows,tfields,ttype,tform,tunit,
extname,&status);
 
/* Read column values from the first extension of the input file and write
   them to a column in the binary extension of the output file created above */
 
tmpanode = (char*) malloc(16*sizeof(char));
 
if((getindex(tevtb_out,"C")) != -1){
        pos=getindex(tevtb_out,"T");
        tmpstr[0]=tevtb_out[pos-3];
        tmpstr[1]='\0';
        cbitnum=atoi(tmpstr);

	printf("\nThere are %d bits for channels.\n",cbitnum);

k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TDOUBLE, 1, i+1, felem, 1,
                &dblnull, &time1, &anynull, &status);
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);
 
        if(event[0] > 127 ){
                temp = event[0] >> 4;  /*get rid of the last 4 bits so that the
                                   remaining bits only have infomation
                                   about M and D tokens */

                temp1=event[0]-(temp*16); /*get the last 4 bits in event[0] */
 
                event[0]=(event[0] >> 4) - 8;  /*make sure it is a real 
                                                event (M token value
                                   equals 1), and get rid of the 1st bit (M)
                                   so that the remaining bits only have
                                   information about detectors (D token)  */
 
		/* get the 7 bits in C token */
                event[1]=temp1*pow(2,(cbitnum-4))+(event[1]>>(12-cbitnum)); 
		tmpanode = "INDEF";

                fits_write_col(outptr,TDOUBLE,1,k+1,felem,1,&time1,&status);
                fits_write_col(outptr,TINT,2,k+1,felem,1,&event[1],&status);
                fits_write_col(outptr,TINT,3,k+1,felem,1,&event[0],&status);
                fits_write_col(outptr,TSTRING,4,k+1,felem,1,&tmpanode,&status);
                k++;
        }
}

}else{		/*if (getindex(tevtb_out,"C") == -1) */

printf("\nThere are no channel bits.\n");

k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TDOUBLE, 1, i+1, felem, 1,
                &dblnull, &time1, &anynull, &status);
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);
 
        if(event[0] > 127 ){
                temp = event[0] >> 4;  /*get rid of the last 4 bits so that the
                                   remaining bits only have infomation
                                   about M and D tokens */
 
                event[0]=(event[0] >> 4) - 8;  /*make sure it is a real
                                                event (M token value
                                   equals 1), and get rid of the 1st bit (M)
                                   so that the remaining bits only have
                                   information about detectors (D token)  */
 
		event[1]=-999;
		tmpanode="INDEF";

                fits_write_col(outptr,TDOUBLE,1,k+1,felem,1,&time1,&status);
         	fits_write_col(outptr,TINT,2,k+1,felem,1,&event[1],&status); 
                fits_write_col(outptr,TINT,3,k+1,felem,1,&event[0],&status);
                fits_write_col(outptr,TSTRING,4,k+1,felem,1,&tmpanode,&status);
                k++;
        }
}
} 

fits_modify_comment(outptr,"TTYPE2",
        "label for field   2 (relative channels)",&status);
 
}  /*end of the 3rd IF */

if((strncmp(keyname,"E_1us_4",7) == 0) || (strncmp(keyname,"E_2us_8",7) == 0) || (strncmp(keyname,"E_4us_16",8) == 0) || (strncmp(keyname,"E_8us_32",8) == 0) || (strncmp(keyname,"E_16us_64",9) == 0)){  /*The 4th IF statement*/
 
/* Read EVENT column from the binary extension of the input file, select rows
   that have M token value of 1, i.e., the first element in each row of EVENT
   colum should be greater than 127.  Record the total number of such rows,
   this number will be used as the row number in the output file.
*/
 
k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);
        if(event[0] > 127){
                k++;
        }
}
nrows=k;
printf("\nThere are %ld events corresponding to M token value equal to 1.\n",nrows);
 
/* Create a binary extension in the output file created above */
 
tfields=4;

fits_create_tbl(outptr,BINARY_TBL,nrows,tfields,ttype,tform,tunit,
extname,&status);
 
/* Read column values from the first extension of the input file and write
   them to a column in the binary extension of the output file created above */
 
tmpanode = (char*) malloc(16*sizeof(char));

if((getindex(tevtb_out,"C")) != -1){
        pos=getindex(tevtb_out,"T");
        tmpstr[0]=tevtb_out[pos-3];
        tmpstr[1]='\0';
        cbitnum=atoi(tmpstr);

	printf("\nThere are %d bits for channels.\n",cbitnum);
 
k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TDOUBLE, 1, i+1, felem, 1,
                &dblnull, &time1, &anynull, &status);
        fits_read_col(fptr, TINT, 2, i+1, felem, 2,
                &longnull, event, &anynull, &status);
 
        if(event[0] > 127 ){
                temp = event[0] >> (7-cbitnum);  /*get rid of the last bit 
						so that the
                                   remaining bits only have infomation
                                   about M and C tokens */
 
                event[0]=temp - pow(2,cbitnum);  /*make sure it is a real 
                                                event (Mtoken value
                                   equals 1), and get rid of the 1st bit (M)
                                   so that the remaining bits only have
                                   information about channels (C token)  */
 
		pcuid=-999;
		tmpanode = "INDEF";
 
                fits_write_col(outptr,TDOUBLE,1,k+1,felem,1,&time1,&status);
                fits_write_col(outptr,TINT,2,k+1,felem,1,&event[0],&status);
                fits_write_col(outptr,TINT,3,k+1,felem,1,&pcuid,&status);
                fits_write_col(outptr,TSTRING,4,k+1,felem,1,&tmpanode,&status);
                k++;
        }
}

}else{
	printf("No C token, exit! \n");
	exit(1);
}
 
fits_modify_comment(outptr,"TTYPE2",
        "label for field   2 (relative channels)",&status);
 
}  /*end of the 4th IF */

if((strncmp(keyname,"GoodXenon",9) == 0)){  /*The 5th IF statement*/

/* Read EVENT column from the binary extension of the input file, select rows
   that have M token value of 1, i.e., the first element in each row of EVENT
   colum should be greater than 127.  Record the total number of such rows,
   this number will be used as the row number in the output file.
*/
k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TINT, 2, i+1, felem, 3,
                &longnull, evt, &anynull, &status);
        if(evt[0] > 127){
                k++;
        }
}
nrows=k;
printf("\nThere are %ld events corresponding to M token value equal to 1.\n",nrows);
 
/* Create a binary extension in the output file created above */
 
tfields=5;
fits_create_tbl(outptr,BINARY_TBL,nrows,tfields,ttype1,tform1,tunit1,
extname,&status);

/* Read column values from the first extension of the input file and write
   them to a column in the binary extension of the output file created above */
 
tmpanode = (char*) malloc(16*sizeof(char));
 
k=0;
for(i=0;i<naxes2;i++){
        fits_read_col(fptr, TDOUBLE, 1, i+1, felem, 1,
                &dblnull, &time1, &anynull, &status);
        fits_read_col(fptr, TINT, 2, i+1, felem, 3,
                &longnull, evt, &anynull, &status);
 
        if(evt[0] > 127){
		temp=evt[0]>>1;
		temp1=temp>>1;
		propane=temp - temp1*2;  /*obtain PROPANE bit*/

		temp1=evt[0] - temp*2;  /*get the first bit for D token*/
		pcuid=temp1*4 + (evt[1]>>6); /*get PCUID*/

		temp=evt[1]>>6;
		temp1=evt[1]-temp*64;  /*get the 6 bits for E token*/

		switch(temp1){   /* determine E token (ANODE) value */
			case 1: 
				strcpy(anode,"X1L");
				break;
			case 2: 
				strcpy(anode,"X1R");
				break;
			case 4: 
				strcpy(anode,"X2L");
				break;
			case 8: 
				strcpy(anode,"X2R");
				break;
			case 16: 
				strcpy(anode,"X3L");
				break;
			case 32: 
				strcpy(anode,"X3R");
				break;
			default: 
				strcpy(anode,"UNKNOWN");
			}

		strcpy(tmpanode,anode);

                fits_write_col(outptr,TDOUBLE,1,k+1,felem,1,&time1,&status);
                fits_write_col(outptr,TINT,2,k+1,felem,1,&evt[2],&status);
                fits_write_col(outptr,TINT,3,k+1,felem,1,&pcuid,&status);
                fits_write_col(outptr,TSTRING,4,k+1,felem,1,&tmpanode,&status);
		fits_write_col(outptr,TINT,5,k+1,felem,1,&propane,&status);
                k++;
        }
}
 
fits_modify_comment(outptr,"TTYPE2",
        "label for field   2 (relative channels)",&status);
 
}       /*end of the 5th IF statement*/


fits_write_key(outptr,TSTRING,"TDDES2",tddes,comm,&status);
fits_write_key_longstr(outptr,"TEVTB2",tevtb_out,comm,&status);
/* (MJT) 30May2001: need LONGSTRN keyword */
fits_write_key_longwarn(outptr,&status);

/* Close the files */

fits_close_file(fptr,&status);
fits_close_file(outptr,&status);
}

int getindex(s,t)
char s[],t[];
{
        int i,j,k;      
        for(i=0;s[i] != '\0';i++){      
                for(j=i,k=0;t[k] != '\0' && s[j] == t[k];j++,k++)       
                        ;       
                if(t[k] == '\0')        
                        return(i);      
        }       
        return(-1);     
}

void getinfo(int idnum,int *pcuid,char * anodename) 
{
/*
	anodename=(char*)malloc(8*sizeof(char)); 
*/
	switch(idnum){
		case 0:
			*pcuid=0;
			strcpy(anodename,"X1L");
			break;
                case 1:
                        *pcuid=0;
                        strcpy(anodename,"X1R");
                        break;
                case 2:
                        *pcuid=0;
                        strcpy(anodename,"X2L");
                        break;
                case 3:
                        *pcuid=0;
                        strcpy(anodename,"X2R");
                        break;
                case 4:
                        *pcuid=0;
                        strcpy(anodename,"X3L");
                        break;
                case 5:
                        *pcuid=0;
                        strcpy(anodename,"X3R");
                        break;
                case 6:
                        *pcuid=1;
                        strcpy(anodename,"X1L");
                        break;
                case 7:
                        *pcuid=1;
                        strcpy(anodename,"X1R");
                        break;
                case 8:
                        *pcuid=1;
                        strcpy(anodename,"X2L");
                        break;
                case 9:
                        *pcuid=1;
                        strcpy(anodename,"X2R");
                        break;
                case 10:
                        *pcuid=1;
                        strcpy(anodename,"X3L");
                        break;
                case 11:
                        *pcuid=1;
                        strcpy(anodename,"X3R");
                        break;
                case 12:
                        *pcuid=2;
                        strcpy(anodename,"X1L");
                        break;
                case 13:
                        *pcuid=2;
                        strcpy(anodename,"X1R");
                        break;
                case 14:
                        *pcuid=2;
                        strcpy(anodename,"X2L");
                        break;
                case 15:
                        *pcuid=2;
                        strcpy(anodename,"X2R");
                        break;
                case 16:
                        *pcuid=2;
                        strcpy(anodename,"X3L");
                        break;
                case 17:
                        *pcuid=2;
                        strcpy(anodename,"X3R");
                        break;
                case 18:
                        *pcuid=3;
                        strcpy(anodename,"X1L");
                        break;
                case 19:
                        *pcuid=3;
                        strcpy(anodename,"X1R");
                        break;
                case 20:
                        *pcuid=3;
                        strcpy(anodename,"X2L");
                        break;
                case 21:
                        *pcuid=3;
                        strcpy(anodename,"X2R");
                        break;
                case 22:
                        *pcuid=3;
                        strcpy(anodename,"X3L");
                        break;
                case 23:
                        *pcuid=3;
                        strcpy(anodename,"X3R");
                        break;
                case 24:
                        *pcuid=4;
                        strcpy(anodename,"X1L");
                        break;
                case 25:
                        *pcuid=4;
                        strcpy(anodename,"X1R");
                        break;
                case 26:
                        *pcuid=4;
                        strcpy(anodename,"X2L");
                        break;
                case 27:
                        *pcuid=4;
                        strcpy(anodename,"X2R");
                        break;
                case 28:
                        *pcuid=4;
                        strcpy(anodename,"X3L");
                        break;
                case 29:
                        *pcuid=4;
                        strcpy(anodename,"X3R");
                        break;
		default:
			*pcuid=-999;
			strcpy(anodename,"UNKNOWN");
			break;
	}
/*
printf("Subroutine: idnum=%d,*pcuid=%d,anodename=%s\n",idnum,*pcuid,anodename); 
*/
}

int evt_rd_param(ifile,ofile)
	char *ifile, *ofile;
{
int parstat=0;
int BufLen_2=255;

Uclgst("infile",ifile,&parstat);
if(parstat != 0){
	printf("\nCould not open input file.\n");
	exit(1);
}

Uclgst("outfile",ofile,&parstat);
if(parstat != 0){
        printf("\nCould not open output file.\n");
        exit(1);
}

return parstat;
}

/* This code is needed by IRAF */

#ifdef vms
#define F77CALL decodeev
#endif
#ifdef unix
#define F77CALL decodeev_
#endif

void F77CALL()
{
void decodeevt();
decodeevt();
}
